#!/usr/bin/env python
# coding=utf-8
# Filename: chatbot.py
# Author: Tamas Gal <tgal@km3net.de>
"""
The monitoring chatbot.

Usage:
    chatbot.py
    chatbot.py (-h | --help)

Options:
    -h --help   Show this screen.

"""
from datetime import datetime
from pprint import pprint
import random
import re
import requests
import subprocess
import time
import threading
import toml
from rocketchat_API.rocketchat import RocketChat

import km3pipe as kp

log = kp.logger.get_logger("chatbot")

URL = "https://chat.km3net.de"
CONFIG = "pipeline.toml"
RECONNECT_INTERVAL = 30


class RocketChatBot(object):
    def __init__(self, botname, passwd, server, command_character=None):
        self.botname = botname
        self.api = RocketChat(user=botname, password=passwd, server_url=server)
        self.commands = [(['echo', ], self.echo)]
        self.auto_answers = []
        self.direct_answers = []
        self.unknow_command = ['command not found', ]
        self.lastts = {}
        self.command_character = command_character

    def echo(self, msg, user, channel_id):
        self.send_message('@' + user + ' : ' + msg, channel_id)

    def get_status(self, auser):
        return self.api.users_get_presence(username=auser)

    def send_message(self, msg, channel_id):
        self.api.chat_post_message(channel=channel_id, text=msg)

    def add_dm_handler(self, command, action):
        self.commands.append((command, action))

    def add_auto_answer(self, triggers, answers):
        self.auto_answers.append((triggers, answers))

    def add_direct_answer(self, triggers, answers):
        self.direct_answers.append((triggers, answers))

    def handle_command_character_message(self, message, channel_id):
        msg = message['msg'].lstrip(self.command_character)

        command = msg.split()[0].lower()
        arguments = " ".join(msg.split()[1:])
        user = message['u']['username']

        for cmd_list in self.commands:
                if command.lower() in cmd_list[0]:
                    cmd_list[1](arguments, user, channel_id)
                    return

        if not self.handle_auto_answer(message, self.direct_answers, channel_id):
                self.send_message('@' + user + ' :' + random.choice(self.unknow_command), channel_id)

    def handle_direct_message(self, message, channel_id):
        msg = message['msg'].lstrip('@' + self.botname).strip()
        if len(msg) > 0:
            command = msg.split()[0].lower()
            arguments = " ".join(msg.split()[1:])
            user = message['u']['username']
            for cmd_list in self.commands:
                if command.lower() in cmd_list[0]:
                    cmd_list[1](arguments, user, channel_id)
                    return

            if not self.handle_auto_answer(message, self.direct_answers, channel_id):
                self.send_message('@' + user + ' :' + random.choice(self.unknow_command), channel_id)
        else:
            self.send_message('Here I am', channel_id)

    def handle_auto_answer(self, message, answers, channel_id):
        for kind in answers:
            for k in kind[0]:
                if k in message['msg'].lower():
                    self.send_message(random.choice(kind[1]) + ' @' + message['u']['username'], channel_id)
                    return True
        return False

    def handle_messages(self, messages, channel_id):
        for message in messages['messages']:
            if message['u']['username'] != self.botname:
                pprint(message)
                if message['u']['username'] == 'rocket.cat':
                    continue
                if message['msg'].startswith('@' + self.botname):
                    threading.Thread(target=self.handle_direct_message, args=(message, channel_id)).start()
                elif self.command_character is not None and message['msg'].startswith(self.command_character):
                    threading.Thread(target=self.handle_command_character_message, args=(message, channel_id)).start()
                elif 'mentions' not in message or message.get('mentions') == []:
                    threading.Thread(target=self.handle_auto_answer, args=(message, self.auto_answers, channel_id)).start()

    def load_ts(self, channel_id, messages):
        if len(messages) > 0:
            self.lastts[channel_id] = messages[0]['ts']
        else:
            self.lastts[channel_id] = ''

    def load_channel_ts(self, channel_id):
        self.load_ts(channel_id, self.api.channels_history(channel_id).json()['messages'])

    def load_group_ts(self, channel_id):
        self.load_ts(channel_id, self.api.groups_history(channel_id).json()['messages'])

    def load_im_ts(self, channel_id):
        response = self.api.im_history(channel_id).json()
        if response.get('success'):
            self.load_ts(channel_id, self.api.im_history(channel_id).json()['messages'])

    def process_messages(self, messages, channel_id):
        try:
            if "success" in messages:
                if messages['success'] == False:
                    raise RuntimeError(messages['error'])
            if len(messages['messages']) > 0:
                self.lastts[channel_id] = messages['messages'][0]['ts']
            self.handle_messages(messages, channel_id)
        except Exception as e:
            pprint(e)

    def process_channel(self, channel_id):
        if channel_id not in self.lastts:
            self.lastts[channel_id] = datetime.now().isoformat()

        self.process_messages(self.api.channels_history(channel_id, oldest=self.lastts[channel_id]).json(),
                              channel_id)

    def process_group(self, channel_id):
        if channel_id not in self.lastts:
            self.lastts[channel_id] = ''

        self.process_messages(self.api.groups_history(channel_id, oldest=self.lastts[channel_id]).json(),
                              channel_id)

    def process_im(self, channel_id):
        if channel_id not in self.lastts:
            self.lastts[channel_id] = ''

        self.process_messages(self.api.im_history(channel_id, oldest=self.lastts[channel_id]).json(),
                              channel_id)

    def run(self):
        for channel in self.api.channels_list_joined().json().get('channels'):
            self.load_channel_ts(channel.get('_id'))

        for group in self.api.groups_list().json().get('groups'):
            self.load_group_ts(group.get('_id'))

        for im in self.api.im_list().json().get('ims'):
            self.load_im_ts(im.get('_id'))

        while 1:
            for channel in self.api.channels_list_joined().json().get('channels'):
                threading.Thread(target=self.process_channel, args=(channel.get('_id'),)).start()

#            for group in self.api.groups_list().json().get('groups'):
#                threading.Thread(target=self.process_group, args=(group.get('_id'),)).start()
#
#            for im in self.api.im_list().json().get('ims'):
#                threading.Thread(target=self.process_im, args=(im.get('_id'),)).start()

            time.sleep(1)


with open(CONFIG, 'r') as fobj:
    print(f"Reading configuration from {CONFIG}")
    config = toml.load(fobj)
    BOTNAME = config['Alerts']['botname']
    PASSWORD = config['Alerts']['password']
    CHANNEL = config['Alerts']['channel']


def get_channel_id(channel):
    print(f"Getting channel ID for channel: {channel}")
    tries = 0
    while True:
        tries += 1

        try:
            rocket = RocketChat(BOTNAME, PASSWORD, server_url=URL)
        except requests.exceptions.ConnectionError as e:
            log.error(f"Unable to connect to the RocketChat server: {e}")
        except Exception as e:
            log.error(f"Unknown error occured: {e}")
        else:
            break

        interval = tries * RECONNECT_INTERVAL
        print(f"Reconnecting in {interval} seconds...")
        time.sleep(interval)

    channels = rocket.channels_list(count=0).json()['channels']
    for c in channels:
        print(f"   -> {c['name']} => {c['_id']}")
        if c['name'] == channel:
            print(f"Found channel ID for {channel} is {c['_id']}")
            return c['_id']
    log.error("No channel found with name {}".format(channel))


CHANNEL_ID = get_channel_id(CHANNEL)


def run():
    print("Running the monitoring bot system")
    bot = spawn_bot()
    register_handlers(bot)
#    bot.send_message("ChatBot (re)started. I am up and running!", CHANNEL_ID)
    bot.run()


def spawn_bot():
    print("Spawning the bot")
    return RocketChatBot(BOTNAME, PASSWORD, URL)


def is_shifter(user):
    print(f"Checking if {user} is a shifter")
    with open(CONFIG, 'r') as fobj:
        config = toml.load(fobj)
        try:
            alerts_config = config['Alerts']
        except KeyError:
            log.error("No 'Alerts' section found in the configuration file")
            return False
        try:
            return user in alerts_config['shifters']
        except KeyError:
            log.error("No 'shifters' section found in 'Alerts' of the configuration file")
            return False


def is_operator(user):
    print(f"Checking if {user} is an operator")
    with open(CONFIG, 'r') as fobj:
        config = toml.load(fobj)
        return user in config['Alerts']['operators']


def register_handlers(bot):
    print("Registering API handlers")
    def greet(msg, user, channel_id):
        if channel_id != CHANNEL_ID:
            print("skipping")
            return
        bot.send_message('hello @' + user, channel_id)

    def status(msg, user, channel_id):
        print(f"Reporting status to channel {channel_id}")
        if channel_id != CHANNEL_ID:
            print(f"Skipping channel with ID {channel_id}")
            return
        if not is_shifter(user) and not is_operator(user):
            bot.send_message(
                "Sorry @{}, only operators and shifters are allowed to mess "
                "with me, sorry...".format(user), channel_id)
            return
        print("Asking subservisorctl for the status")
        try:
            status = "```\n" + subprocess.check_output(
                ['supervisorctl', 'status']).decode('ascii') + "\n```"
        except subprocess.CalledProcessError as e:
            status = "```\n{}\n```".format(e.output.decode('ascii'))
        print("Sending status")
        bot.send_message(status, channel_id)

    def supervisorctl(msg, user, channel_id):
        if channel_id != CHANNEL_ID:
            print("skipping")
            return
        if not is_shifter(user) and not is_operator(user):
            bot.send_message(
                "Sorry @{}, only operators and shifters are allowed to mess "
                "with me, sorry...".format(user), channel_id)
            return
        try:
            output = subprocess.check_output(
                ['supervisorctl'] + msg.split(),
                stderr=subprocess.STDOUT).decode('ascii')
        except subprocess.CalledProcessError as e:
            output = e.output.decode('ascii')
        print("supervisorctl ({}), called by {}:\n{}".format(
            msg, user, output))
        bot.send_message(output, channel_id)

    def shifters(msg, user, channel_id):
        if channel_id != CHANNEL_ID:
            print("skipping")
            return
        if not is_operator(user):
            bot.send_message(
                "Sorry @{}, only operators are allowed to set shifters!".
                format(user), channel_id)
            return
        try:
            with open(CONFIG, 'r') as fobj:
                config = toml.load(fobj)
            shifters = msg[3:].strip()
            config['Alerts']['shifters'] = shifters
            with open(CONFIG, 'w') as fobj:
                toml.dump(config, fobj)
            msg = f'Alright, the new shifters are {shifters}, welcome!'
            print(msg)
            bot.send_message(msg, channel_id)
        except Exception as e:
            bot.send_message(f'something went horribly wrong... {e}',
                             channel_id)

    def help(msg, user, channel_id):
        if channel_id != CHANNEL_ID:
            print("skipping", channel_id)
            return
        help_str = f"""
        Hi @{user} I was built to take care of the monitoring alerts.
        Here is how you can use me:
        - `@{BOTNAME} shifters are cnorris and bspencer`
          -> set the new shifters who I may annoy with chat messages and
          emails.
        - `@{BOTNAME} status` -> show the status of the monitoring system
        - `@{BOTNAME} supervisorctl` -> take control over the monitoring system
        - `@{BOTNAME} help` -> show this message
        """

        bot.send_message(help_str, channel_id)

    handlers = [(['hello', 'hey', 'hi', 'ciao'], greet), (['status'], status),
                (['help'], help), (['shifters'], shifters),
                (['supervisorctl'], supervisorctl)]
    for trigger, handler in handlers:
        bot.add_dm_handler(trigger, handler)
    print("All handlers are registered")


def main():
    from docopt import docopt
    args = docopt(__doc__)

    run()


if __name__ == '__main__':
    main()
